from __future__ import annotations

import typing

from sqlalchemy import orm

from lime_uow import exceptions


__all__ = ("SqlAlchemySession",)

from lime_uow.resources import resource


class SqlAlchemySession(resource.Resource[orm.Session]):
    def __init__(self, session_factory: orm.sessionmaker, /):
        self._session_factory = session_factory
        self._session: typing.Optional[orm.Session] = None

    @classmethod
    def interface(cls) -> typing.Type[SqlAlchemySession]:
        return cls

    def open(self) -> orm.Session:
        self._session = self._session_factory()
        return self._session

    def close(self) -> None:
        if self._session:
            self._session.close()
            self._session = None

    def rollback(self) -> None:
        if self._session is None:
            raise exceptions.RollbackError(
                "Attempted to rollback a closed session.",
            )
        else:
            self._session.rollback()

    def save(self) -> None:
        if self._session:
            self._session.commit()



import abc
import typing

from sqlalchemy import orm

from lime_uow.resources.repository import repository

E = typing.TypeVar("E")

__all__ = ("SqlAlchemyRepository",)


class SqlAlchemyRepository(repository.Repository[E], abc.ABC, typing.Generic[E]):
    def __init__(self, session: orm.Session, /):
        self._session = session
        self._entity_type: typing.Optional[typing.Type[E]] = None

    def add(self, item: E, /) -> E:
        self.session.add(item)
        return item

    def add_all(self, items: typing.Collection[E], /) -> typing.Collection[E]:
        self.session.bulk_save_objects(items)
        return items

    def all(self) -> typing.Generator[E, None, None]:
        return self.session.query(self.entity_type).all()

    def delete(self, item: E, /) -> E:
        self.session.delete(item)
        return item

    def delete_all(self) -> None:
        self.session.query(self.entity_type).delete(synchronize_session=False)

    def get(self, item_id: typing.Any, /) -> E:
        return self.session.query(self.entity_type).get(item_id)

    @property
    @abc.abstractmethod
    def entity_type(self) -> typing.Type[E]:
        raise NotImplementedError

    def rollback(self) -> None:
        self.session.rollback()

    def save(self) -> None:
        self.session.commit()

    @property
    def session(self) -> orm.Session:
        return self._session

    def set_all(self, items: typing.Collection[E], /) -> typing.Collection[E]:
        self.session.query(self.entity_type).delete()
        self.session.bulk_save_objects(items)
        return items

    def update(self, item: E, /) -> E:
        self.session.merge(item)
        return item

    def where(self, predicate: typing.Any, /) -> typing.List[E]:
        return self.session.query(self.entity_type).filter(predicate).all()



from __future__ import annotations

import typing

from lime_uow import exceptions, resources

__all__ = (
    "SharedResources",
    "PlaceholderSharedResources",
)

T = typing.TypeVar("T")


class SharedResources:
    def __init__(self, /, *shared_resource: resources.Resource[typing.Any]):
        resources.check_for_ambiguous_implementations(shared_resource)
        self.__shared_resources = tuple(shared_resource)
        self.__handles: typing.Dict[str, typing.Any] = {}
        self.__opened = False
        self.__closed = False

    def __enter__(self) -> SharedResources:
        if self.__opened:
            raise exceptions.ResourcesAlreadyOpen()
        if self.__closed:
            raise exceptions.ResourceClosed()
        self.__opened = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()

    def close(self):
        if self.__closed:
            raise exceptions.ResourceClosed()
        for resource in self.__shared_resources:
            resource.close()
        self.__handles = {}
        self.__closed = True
        self.__opened = False

    def get(
        self,
        resource_type: typing.Type[resources.Resource[T]],
    ) -> T:
        if self.__closed:
            raise exceptions.ResourceClosed()
        if resource_type.interface() in self.__handles.keys():
            return self.__handles[resource_type.interface().__name__]
        else:
            try:
                resource = next(
                    resource
                    for resource in self.__shared_resources
                    if resource.interface() == resource_type.interface()
                )
                handle = resource.open()
                self.__handles[resource.interface().__name__] = handle
                return handle
            except StopIteration:
                raise exceptions.MissingResourceError(
                    resource_name=resource_type.interface().__name__,
                    available_resources={
                        r.interface().__name__ for r in self.__shared_resources
                    },
                )
            except Exception as e:
                raise exceptions.LimeUoWException(str(e))

    def __eq__(self, other: object) -> bool:
        if other.__class__ is self.__class__:
            # noinspection PyTypeChecker
            return (
                self.__shared_resources
                == typing.cast(SharedResources, other).__shared_resources
            )
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return hash(self.__shared_resources)

    def __repr__(self) -> str:
        resources_str = ", ".join(
            r.interface().__name__ for r in self.__shared_resources
        )
        return f"{self.__class__.__name__}: {resources_str}"


class PlaceholderSharedResources(SharedResources):
    def __init__(self):
        super().__init__()



from __future__ import annotations

import typing

import pyodbc

from lime_uow.resources import resource

__all__ = ("PyodbcCursor",)


class PyodbcCursor(resource.Resource[pyodbc.Cursor]):
    def __init__(
        self,
        *,
        con: pyodbc.Connection,
        fast_executemany: bool = True,
    ):
        self._con = con
        self._fast_executemany = fast_executemany

        self._handle: typing.Optional[pyodbc.Cursor] = None

    def close(self) -> None:
        if self._handle is not None:
            self._handle.close()
            self._handle = None

    def open(self) -> pyodbc.Cursor:
        if self._handle is None:
            self._handle = self._con.cursor()
            self._handle.fast_executemany = self._fast_executemany
            # self._handle.setinputsizes([(pyodbc.SQL_WVARCHAR, 0, 0)])
        return self._handle

    @classmethod
    def interface(cls) -> typing.Type[PyodbcCursor]:
        return PyodbcCursor

    def rollback(self) -> None:
        if self._handle is not None:
            self._handle.rollback()

    def save(self) -> None:
        if self._handle is not None:
            self._handle.commit()



from __future__ import annotations

import typing

import pyodbc

from lime_uow import exceptions
from lime_uow.resources import resource

__all__ = ("PyodbcConnection",)


class PyodbcConnection(resource.Resource[pyodbc.Connection]):
    def __init__(
        self,
        db_uri: str,
        autocommit: bool = False,
        read_only: bool = False,
    ):
        self._db_uri = db_uri
        self._autocommit = autocommit
        self._read_only = read_only

        self._handle: typing.Optional[pyodbc.Connection] = None

    def open(self) -> pyodbc.Connection:
        if self._handle is None:
            self._handle = pyodbc.connect(
                self._db_uri,
                autocommit=self._autocommit,
                readonly=self._read_only,
            )
            self._handle.maxwrite = 1024 * 1024 * 1024
        return self._handle

    def close(self) -> None:
        if self._handle is not None:
            self._handle.close()
            self._handle = None

    @classmethod
    def interface(cls) -> typing.Type[PyodbcConnection]:
        return PyodbcConnection

    def rollback(self) -> None:
        if self._handle is not None:
            self._handle.rollback()

    def save(self) -> None:
        if self._read_only:
            raise exceptions.ReadOnlyConnection(
                "Attempted to call save() on a read-only connection."
            )
        if self._handle is None:
            raise exceptions.ResourceClosed()
        else:
            self._handle.commit()



from __future__ import annotations

import typing

import sqlalchemy as sa

from lime_uow.resources.resource import Resource

__all__ = ("SqlAlchemyTransaction",)


class SqlAlchemyTransaction(Resource[sa.engine.Transaction]):
    def __init__(self, /, engine: sa.engine.Engine):
        self._engine = engine
        self._transaction: typing.Optional[sa.engine.Transaction] = None

    def open(self) -> sa.engine.Transaction:
        if self._transaction is None:
            self._transaction = self._engine.begin()
        return self._transaction

    def close(self) -> None:
        if self._transaction is not None:
            self._transaction.close()
            self._transaction = None

    @classmethod
    def interface(cls) -> typing.Type[SqlAlchemyTransaction]:
        return SqlAlchemyTransaction

    def rollback(self) -> None:
        if self._transaction is not None:
            self._transaction.rollback()

    def save(self) -> None:
        if self._transaction is not None:
            self._transaction.commit()



from __future__ import annotations

import typing

import sqlalchemy as sa

from lime_uow.resources import resource, sqlalchemy_transaction

__all__ = ("SqlAlchemyEngine",)


class SqlAlchemyEngine(resource.Resource[sa.engine.Engine]):
    def __init__(self, /, db_uri: str):
        self._db_uri = db_uri
        self._engine: typing.Optional[sa.engine.Engine] = None

    def transaction(self) -> sqlalchemy_transaction.SqlAlchemyTransaction:
        return sqlalchemy_transaction.SqlAlchemyTransaction(self._engine)

    def open(self) -> sa.engine.Engine:
        if self._engine is None:
            self._engine = sa.create_engine(self._db_uri)
        return self._engine

    def close(self) -> None:
        pass

    @classmethod
    def interface(cls) -> typing.Type[SqlAlchemyEngine]:
        return SqlAlchemyEngine

    def rollback(self) -> None:
        pass

    def save(self) -> None:
        pass



from __future__ import annotations

import os
import pathlib
import tempfile
import typing

from lime_uow.resources import resource

__all__ = ("TempFileSharedResource",)


class TempFileSharedResource(resource.Resource[typing.IO[bytes]]):
    def __init__(
        self,
        *,
        prefix: typing.Optional[str] = None,
        file_extension: typing.Optional[str] = None,
    ):
        self._prefix = prefix
        self._file_extension = file_extension

        self._file_handle: typing.Optional[typing.IO[bytes]] = None
        self._file_path: typing.Optional[pathlib.Path] = None

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def clear(self):
        self.handle.seek(0, 0)  # go to beginning of file
        self.handle.truncate()

    def close(self) -> None:
        self.handle.close()
        os.unlink(self.file_path)

    @property
    def handle(self) -> typing.IO[bytes]:
        if self._file_handle is None:
            self.open()
        return self._file_handle  # type: ignore

    @classmethod
    def interface(cls) -> typing.Type[TempFileSharedResource]:
        return cls

    @property
    def file_path(self) -> pathlib.Path:
        if self._file_path is None:
            self._file_path = pathlib.Path(self.handle.name)
        return self._file_path

    def open(self) -> typing.IO[bytes]:
        ext = f".{self._file_extension}" if self._file_extension else None
        self._file_handle = tempfile.NamedTemporaryFile(
            prefix=self._prefix,
            suffix=ext,
            delete=False,
        )
        return self._file_handle

    def all(self) -> str:
        self.handle.seek(0, 0)  # go to beginning of file
        content = self.handle.read()
        self.handle.seek(0, 2)  # go to end of file
        return content.decode()

    def rollback(self) -> None:
        pass

    def save(self) -> None:
        self.handle.flush()

    def add(self, content: str) -> None:
        self.handle.write(content.encode())



import abc
import typing

from lime_uow.resources import resource

E = typing.TypeVar("E")


__all__ = ("Repository",)


class Repository(resource.Resource[E], abc.ABC, typing.Generic[E]):
    """Interface to access elements of a collection"""

    @abc.abstractmethod
    def add(self, item: E, /) -> E:
        raise NotImplementedError

    @abc.abstractmethod
    def add_all(self, items: typing.Collection[E], /) -> typing.Collection[E]:
        raise NotImplementedError

    @abc.abstractmethod
    def all(self) -> typing.Iterable[E]:
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, item: E, /) -> E:
        raise NotImplementedError

    @abc.abstractmethod
    def delete_all(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def set_all(self, items: typing.Collection[E], /) -> typing.Collection[E]:
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, item: E, /) -> E:
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, item_id: typing.Any, /) -> E:
        raise NotImplementedError



import typing

__all__ = (
    "LimeUoWException",
    "MultipleRegisteredImplementations",
    "InvalidResource",
    "MissingResourceError",
    "OutsideTransactionError",
    "RollbackError",
)


class LimeUoWException(Exception):
    """Base class for exceptions arising from the lime-uow package"""

    def __init__(self, /, message: str):
        self.message = message
        super().__init__(message)


class MultipleRegisteredImplementations(LimeUoWException):
    def __init__(self, duplicates: typing.Mapping[str, int], /):
        self.duplicates = duplicates

        examples = ", ".join(
            f"{name} = {ct}" for name, ct in sorted(duplicates.items())
        )
        msg = f"Resource names must be unique, but found the following duplicates: {examples}."
        super().__init__(msg)


class InvalidResource(LimeUoWException):
    def __init__(self, /, message: str):
        super().__init__(message)


class MissingResourceError(LimeUoWException):
    def __init__(
        self, *, resource_name: str, available_resources: typing.Iterable[str]
    ):
        self.resource_name = resource_name
        msg = (
            f"Could not locate a resource named {resource_name!r}.  "
            f"Available resources include the following: {', '.join(available_resources)}."
        )
        super().__init__(msg)


class NoCommonAncestor(LimeUoWException):
    def __init__(self, /, message: str):
        super().__init__(message)


class OutsideTransactionError(LimeUoWException):
    def __init__(self):
        super().__init__(
            "Attempted to access a UnitOfWork resource outside a with block."
        )


class ReadOnlyConnection(LimeUoWException):
    def __init__(self, /, message: str):
        super().__init__(message)


class RollbackError(LimeUoWException):
    def __init__(self, /, message: str):
        super().__init__(message)


class RollbackErrors(LimeUoWException):
    def __init__(self, *rollback_errors: RollbackError):
        self.rollback_errors = rollback_errors
        err_msg = (
            f"The following errors occurred while performing a rollback: "
            f"{'; '.join(e.message for e in rollback_errors)}."
        )
        super().__init__(err_msg)


class ResourcesAlreadyOpen(LimeUoWException):
    def __init__(self):
        super().__init__("SharedResources are already open.")


class ResourceClosed(LimeUoWException):
    def __init__(self):
        super().__init__("Attempted to access a closed resource.")



import abc
import typing

from lime_uow.resources.repository import repository

E = typing.TypeVar("E")

__all__ = ("DummyRepository",)


class DummyRepository(repository.Repository[E], abc.ABC, typing.Generic[E]):
    """Repository implementation based on a dictionary

    This exists primarily to make testing in client code simpler.  It was not designed for efficiency.
    """

    def __init__(
        self,
        *,
        key_fn: typing.Callable[[E], typing.Hashable],
        initial_values: typing.Optional[typing.Iterable[E]] = None,
    ):
        super().__init__()

        self._current_state: typing.List[E] = list(initial_values or [])
        self._previous_state: typing.List[E] = self._current_state.copy()
        self._key_fn = key_fn

        self.events: typing.List[typing.Tuple[str, typing.Dict[str, typing.Any]]] = []

    def rollback(self) -> None:
        self.events.append(("rollback", {}))
        self._current_state = self._previous_state.copy()

    def save(self) -> None:
        self.events.append(("save", {}))
        self._previous_state = self._current_state.copy()

    def add(self, item: E, /) -> E:
        self.events.append(("add", {"item": item}))
        self._current_state.append(item)
        return item

    def add_all(self, items: typing.Collection[E], /) -> typing.Collection[E]:
        self.events.append(("add_all", {"items": items}))
        self._current_state += items
        return items

    def all(self) -> typing.Iterable[E]:
        self.events.append(("all", {}))
        return list(self._current_state)

    def delete(self, item: E, /) -> E:
        self.events.append(("delete", {"item": item}))
        self._current_state = [
            o for o in self._current_state if self._key_fn(item) != self._key_fn(o)
        ]
        return item

    def delete_all(self) -> None:
        self.events.append(("delete_all", {}))
        self._current_state = []

    def set_all(self, items: typing.Collection[E], /) -> typing.Collection[E]:
        self.events.append(("set_all", {"items": items}))
        self._current_state = list(items)
        return items

    def update(self, item: E, /) -> E:
        self.events.append(("update", {"item": item}))
        original_index = next(
            ix
            for ix, o in enumerate(self._current_state)
            if self._key_fn(item) == self._key_fn(o)
        )
        self._current_state[original_index] = item
        return item

    def get(self, item_id: typing.Any, /) -> E:
        self.events.append(("get", {"item_id": item_id}))
        return next(o for o in self._current_state if self._key_fn(o) == item_id)



from lime_uow.resources.resource import *
from lime_uow.resources.pyodbc_connection import *
from lime_uow.resources.pyodbc_cursor import *
from lime_uow.resources.sqlalchemy_engine import *
from lime_uow.resources.sqlalchemy_session import *
from lime_uow.resources.sqlalchemy_transaction import *
from lime_uow.resources.temp_file import *
from lime_uow.resources.sqlalchemy_session import *
from lime_uow.resources.repository import *



from __future__ import annotations

import abc
import typing

from lime_uow import exceptions

__all__ = (
    "Resource",
    "check_for_ambiguous_implementations",
)

T = typing.TypeVar("T", covariant=True)


class Resource(abc.ABC, typing.Generic[T]):
    @abc.abstractmethod
    def close(self) -> None:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def interface(cls) -> typing.Type[Resource[T]]:
        raise NotImplementedError

    @abc.abstractmethod
    def open(self) -> T:
        raise NotImplementedError

    @abc.abstractmethod
    def rollback(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def save(self) -> None:
        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        if other.__class__ is self.__class__:
            # noinspection PyTypeChecker
            return (
                self.__class__.interface()
                == typing.cast(Resource[typing.Any], other).__class__.interface()
            )
        else:
            return NotImplemented

    def __ne__(self, other: object) -> bool:
        result = self.__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        else:
            return not result

    def __hash__(self) -> int:
        return hash(self.__class__.interface())

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}: {self.__class__.interface()}"


def check_for_ambiguous_implementations(
    rs: typing.Iterable[Resource[typing.Any]], /
) -> None:
    names = [r.__class__.interface().__name__ for r in rs]
    duplicate_names = {name: ct for name in names if (ct := names.count(name)) > 1}
    if duplicate_names:
        raise exceptions.MultipleRegisteredImplementations(duplicate_names)
